''*******************************
''*  PS/2 Keyboard Driver v2.0  *
''*  (C) 2004 Parallax, Inc.    *
''*  Modified by Michael Green  *
''*   to support Hydra as well  *
''*******************************
{
Modified by Michael Park for Sphinx
2009/06/15
Modified by Dave Hein for spinsim
2010/12/24
}

obj
  ser : "conio"

dat
  keyval long 0

PUB start(pingroup) : okay

'' Start keyboard driver - starts a cog
'' returns false if no cog available
''
''   pingroup = 1st pin number for 2-pin group for PS/2 I/O
''              (1st - data signal, 2nd - clock signal)
''              Even number selects Demo Board I/O convention
''
''     use 100-ohm resistors between pins and jack
''     use 10K-ohm resistors to pull jack-side signals to VDD
''     connect jack-power to 5V, jack-gnd to VSS
''
''   pingroup = 1st pin number for 4-pin group for PS/2 I/O
''              Odd number selects Hydra I/O convention
''
''   1st pin drives NPN base for PS/2 'data' signal pull-down
''   2nd pin reads PS/2 'data' signal
''   3rd pin drives NPN base for PS/2 'clock' signal pull-down
''   4th pin reads PS/2 'clock' signal
''
''   use 2.2K-ohm resistors between 1st/3rd pins and NPN bases
''   use 22K-ohm resistors between 2nd/4th pins and PS/2-side signals
''   use 2.2K-ohm resistors to pull PS/2-side signals to 5V
''   connect PS/2 power to 5V, PS/2 gnd to vss
''
'' all lock-keys will be enabled, NumLock will be initially 'on',
'' and auto-repeat will be set to 15cps with a delay of .5s

  okay := startx(pingroup, %0_000_100, %01_01000)

PUB startx(pingroup, locks, auto) | okay

'' Like start, but allows you to specify lock settings and auto-repeat
''
''   locks = lock setup
''           bit 6 disallows shift-alphas (case set soley by CapsLock)
''           bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
''           bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
''           (eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
''
''   auto  = auto-repeat setup
''           bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
''           bits 4..0 specify repeat rate (0=30cps..31=2cps)
''           (eg %01_00000 = .5s delay, 30cps repeat)

  okay := 1

PUB peekkey

'' Returns a 0 if the buffer is empty
'' returns next key if buffer not empty
'' but doesn't remove the key from buffer

  ifnot keyval
    keyval := ser.rxcheck
    if keyval == -1
      keyval~
  if keyval == 8
    keyval := $c8
  return keyval


PUB key : keycode

'' Get key (never waits)
'' returns key (0 if buffer empty)
  if keyval
    keycode := keyval
    keyval~
  else
    keycode := ser.rxcheck
    if keycode == -1
      keycode~
  if keycode == 8
    keycode := $c8


PUB getkey : keycode

'' Get next key (may wait for keypress)
'' returns key
  if keyval
    keycode := keyval
    keyval~
  else
    keycode := ser.rx
  if keycode == 8
    keycode := $c8

{{
+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  |
+------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    |
|files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}